from .dscnet_2d import DSCNet as DSCNet_2d, DSCNet_pro as DSCNet_2d_pro
from .dscnet_3d import DSCNet as DSCNet_3d, DSCNetSmall as DSCNet_3d_small

# This folder contains code adapted from:
# https://github.com/YaoleiQi/DSCNet
# paper
# https://arxiv.org/abs/2307.08388